@@ -8,35 +8,68 @@
 *
 ***********************************************************************/
 
-static char const RCSID[] =
-"$Id$";
-
 #ifdef EMBED_PERL
 #include <EXTERN.h>
 #include <perl.h>
 #include <errno.h>
 #include <syslog.h>
 
+#define PERLPARSE_NUM_ARGS 6
+
 static PerlInterpreter *my_perl = NULL;
 extern void xs_init ();
 
+void
+init_embedded_interpreter(int argc, char **argv, char **env)
+{
+#ifdef PERL_SYS_INIT3
+    PERL_SYS_INIT3(&argc, &argv, &env);
+#endif
+}
+
+void
+term_embedded_interpreter(void)
+{
+    if (my_perl != NULL) {
+	PL_perl_destruct_level = 1;
+	perl_destruct(my_perl);
+	perl_free(my_perl);
+#ifdef PERL_SYS_TERM
+	PERL_SYS_TERM();
+#endif
+	my_perl = NULL;
+    }
+}
+
+static char **argv = NULL;
+
 int
 make_embedded_interpreter(char const *progPath,
 			  char const *subFilter,
 			  int wantStatusReports,
 			  char **env)
 {
-    char *argv[6];
     int argc;
 
+    /* Why do we malloc argv instead of making it static?  Because on some
+       systems, Perl makes horrendously evil assumptions about the alignment
+       of argv... we use malloc to get guaranteed worst-case alignment.
+       Yes, the Perl innards are completely horrible. */
+    if (!argv) {
+	argv = (char **) malloc(PERLPARSE_NUM_ARGS * sizeof(char *));
+	if (!argv) {
+	    fprintf(stderr, "Out of memory allocating argv[] array for embedded Perl!");
+	    syslog(LOG_ERR, "Out of memory allocating argv[] array for embedded Perl!");
+	    exit(EXIT_FAILURE);
+	}
+    }
+    memset(argv, 0, PERLPARSE_NUM_ARGS * sizeof(char *));
+
     if (my_perl != NULL) {
 #ifdef SAFE_EMBED_PERL
 	PL_perl_destruct_level = 1;
 	perl_destruct(my_perl);
 	perl_free(my_perl);
-#ifdef PERL_SYS_TERM
-	PERL_SYS_TERM();
-#endif
 	my_perl = NULL;
 #else
 	syslog(LOG_WARNING, "Cannot destroy and recreate a Perl interpreter safely on this platform.  Filter rules will NOT be reread.");
@@ -68,9 +101,6 @@ make_embedded_interpreter(char const *progPath,
 	argv[3] = NULL;
 	argc = 3;
     }
-#ifdef PERL_SYS_INIT3
-    PERL_SYS_INIT3(&argc, &argv, &env);
-#endif
 
     my_perl = perl_alloc();
     if (!my_perl) {
