@@ -15,9 +15,6 @@
 * libmilter/README in the Sendmail 8.11 distribution.
 ***********************************************************************/
 
-static char const RCSID[] =
-"$Id$";
-
 #define _BSD_SOURCE 1
 
 #ifdef __linux__
@@ -26,6 +23,8 @@ static char const RCSID[] =
 #define _POSIX_SOURCE 1
 #endif
 
+#define MAX_ML_LINES 31
+
 #ifdef HAVE_SOCKLEN_T
 typedef socklen_t md_socklen_t;
 #else
@@ -152,6 +151,7 @@ static char *EquivToLoopback = NULL;
 static char ValidateHeader[256];
 
 /* Additional Sendmail macros to pass along */
+/// TODO Make the limit 256
 #define MAX_ADDITIONAL_SENDMAIL_MACROS 32
 static char *AdditionalMacros[MAX_ADDITIONAL_SENDMAIL_MACROS];
 static int NumAdditionalMacros = 0;
@@ -207,18 +207,12 @@ static int set_reply(SMFICTX *ctx, char const *first, char const *code, char con
 
 #define DATA ((struct privdata *) smfi_getpriv(ctx))
 
-/* How many tries to get a unique directory name before giving up? */
-#define MAXTRIES 8192
-
 /* Size of chunk when replacing body */
 #define CHUNK 4096
 
 /* Number of file descriptors to close when forking */
 #define CLOSEFDS 256
 
-/* Mutex to protect mkdir() calls */
-static pthread_mutex_t MkdirMutex = PTHREAD_MUTEX_INITIALIZER;
-
 /* Do relay check? */
 static int doRelayCheck = 0;
 
@@ -234,9 +228,6 @@ static int doRecipientCheck = 0;
 /* Keep directories around if multiplexor fails? */
 static int keepFailedDirectories = 0;
 
-/* Protect mkdir() with mutex? */
-static int protectMkdirWithMutex = 0;
-
 /* Number of scanning slaves reserved for connection from loopback */
 static int slavesReservedForLoopback = -1;
 
@@ -315,6 +306,68 @@ put_fd(int fd)
     return -1;
 }
 
+/**********************************************************************
+* %FUNCTION: do_reply
+* %ARGUMENTS:
+*  ctx -- filter context
+*  code -- SMTP three-digit code
+*  dsn -- SMTP DSN status notification code
+*  reply -- text message (MAY BE MODIFIED!)
+* %RETURNS:
+*  Nothing
+* %DESCRIPTION:
+*  Called if we KNOW reply contains carriage-returns.
+*  Sets the SMTP reply code and message.  If smfi_setmlreply is available,
+*  call it.  Otherwise, call smfi_setreply after changing carriage-returns
+*  to spaces.
+***********************************************************************/
+static sfsistat
+do_reply(SMFICTX *ctx, char const *code, char const *dsn, char *reply)
+{
+    char *s;
+    char *lines[MAX_ML_LINES+1];
+    int i;
+
+    /* If there are carriage returns and we don't have smfi_setmlreply,
+       change them to spaces and call smfi_setreply */
+#ifndef MILTER_BUILDLIB_HAS_SETMLREPLY
+    for (s = reply; *s; s++) {
+	if (*s == '\n') *s = ' ';
+    }
+    return smfi_setreply(ctx, (char *) code, (char *) dsn, reply);
+#else
+    /* Split on carriage-returns and pass to smfi_setmlreply */
+    for (i=0; i<MAX_ML_LINES+1; i++) {
+	lines[i] = NULL;
+    }
+    lines[0] = reply;
+    s = reply;
+    for (i=1; i<MAX_ML_LINES; i++) {
+	s = strchr(s, '\n');
+	if (!s) break;
+	*s++ = 0;
+	if (!*s) break;
+	lines[i] = s;
+    }
+
+    /* Convert remaining newlines to spaces */
+    while(s && *s) {
+	if (*s == '\n') *s = ' ';
+	s++;
+    }
+
+    /* Sigh... wtf were milter developers thinking??? */
+    return smfi_setmlreply(ctx, (char *) code, (char *) dsn,
+			   lines[0], lines[1], lines[2], lines[3],
+			   lines[4], lines[5], lines[6], lines[7],
+			   lines[8], lines[9], lines[10], lines[11],
+			   lines[12], lines[13], lines[14], lines[15],
+			   lines[16], lines[17], lines[18], lines[19],
+			   lines[20], lines[21], lines[22], lines[23],
+			   lines[24], lines[25], lines[26], lines[27],
+			   lines[28], lines[29], lines[30], lines[31]);
+#endif
+}
 /**********************************************************************
 * %FUNCTION: set_reply
 * %ARGUMENTS:
@@ -328,7 +381,7 @@ put_fd(int fd)
 * %DESCRIPTION:
 *  Sets the SMTP reply code and message.  code and dsn are validated.
 ***********************************************************************/
-static int
+static sfsistat
 set_reply(SMFICTX *ctx,
 	  char const *first,
 	  char const *code,
@@ -336,6 +389,7 @@ set_reply(SMFICTX *ctx,
 	  char const *reply)
 {
     char *safe_reply;
+    sfsistat r;
     if (!reply || !*reply) {
 	if (*first == '4') {
 	    reply = "Please try again later";
@@ -354,15 +408,14 @@ set_reply(SMFICTX *ctx,
 
     /* We need to double any "%" chars in reply */
     if (strchr(reply, '%')) {
-	int retcode;
+	sfsistat retcode;
 	char const *s;
 	char *t;
 	/* Worst-case, we'll double our length */
 	safe_reply = malloc(strlen(reply) * 2 + 1);
 	if (!safe_reply) {
 	    syslog(LOG_ERR, "Out of memory to escape reply %s", reply);
-	    return smfi_setreply(ctx, (char *) code, (char *) dsn,
-				 "Out of memory");
+	    return smfi_setreply(ctx, (char *) code, (char *) dsn, "Out of memory");
 	}
 	s = reply;
 	t = safe_reply;
@@ -371,13 +424,27 @@ set_reply(SMFICTX *ctx,
 	    *t++ = *s++;
 	}
 	*t = 0;
-	retcode = smfi_setreply(ctx, (char *) code, (char *) dsn, safe_reply);
+	if (!strchr(reply, '\n')) {
+	    retcode = smfi_setreply(ctx, (char *) code, (char *) dsn, safe_reply);
+	} else {
+	    retcode = do_reply(ctx, code, dsn, safe_reply);
+	}
 	free(safe_reply);
 	return retcode;
     }
 
-    /* smfi_setreply is not const-correct, hence the (char *) casts */
-    return smfi_setreply(ctx, (char *) code, (char *) dsn, (char *) reply);
+    /* If there are no carriage-returns, avoid allocating memory */
+    if (!strchr(reply, '\n')) {
+	return smfi_setreply(ctx, (char *) code, (char *) dsn, (char *) reply);
+    }
+
+    safe_reply = strdup_with_log(reply);
+    if (!safe_reply) {
+	return smfi_setreply(ctx, (char *) code, (char *) dsn, "Out of memory");
+    }
+    r = do_reply(ctx, code, dsn, safe_reply);
+    free(safe_reply);
+    return r;
 }
 
 /**********************************************************************
@@ -521,6 +588,22 @@ mfconnect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *sa)
 #ifdef AF_INET6
         if (sa->sa_family == AF_INET6) {
 	    tmp = inet_ntop(AF_INET6, &in6sa->sin6_addr, data->hostip, 65);
+	    /* Convert IPv6-mapped IPv4 address to pure IPv4.  That is:
+	       ::ffff:xxx.yyy.zzz.www to simply xxx.yyy.zzz.www */
+	    if (tmp) {
+		if (IN6_IS_ADDR_V4MAPPED(&in6sa->sin6_addr) ||
+		    IN6_IS_ADDR_V4COMPAT(&in6sa->sin6_addr)) {
+		    if (strchr(data->hostip, '.')) {
+			char const *lastcolon = strrchr(data->hostip, ':');
+			char *dst = data->hostip;
+			while(lastcolon) {
+			    lastcolon++;
+			    *dst++ = *lastcolon;
+			    if (!*lastcolon) break;
+			}
+		    }
+		}
+	    }
 	} else
 #endif
         if (sa->sa_family == AF_INET) {
@@ -680,17 +763,15 @@ static sfsistat
 envfrom(SMFICTX *ctx, char **from)
 {
     struct privdata *data;
-    int tries;
     int i;
-    int success;
     char buffer[SMALLBUF];
     char buf2[SMALLBUF];
-    time_t now = time(NULL);
-    unsigned long ulnow = (unsigned long) now;
     char *queueid;
     char *me;
     dynamic_buffer dbuf;
 
+    char mxid[MX_ID_LEN+1];
+
     DEBUG_ENTER("envfrom");
     /* Get the private context */
     data = DATA;
@@ -734,53 +815,18 @@ envfrom(SMFICTX *ctx, char **from)
 	data->firstRecip = NULL;
     }
 
-    /* We need a nice clean directory all our own.  If two
-     * threads get the same name, one of the mkdir() calls will fail.
-     * On some systems, we might need to lock this region, hence
-     * the MkdirMutex */
-
-    success = 0;
-
     /* Try making a directory whose name contains the Sendmail qid first */
-    if (data->qid && data->qid != NOQUEUE) {
-	snprintf(buffer, SMALLBUF, "%s/mdefang-%s", SpoolDir, data->qid);
-	if (!mkdir(buffer, 0750)) {
-	    success = 1;
-	}
+    if (!data->qid || data->qid == NOQUEUE) {
+        // We have no queue ID, so fail.
+	syslog(LOG_WARNING, "No queue ID available in envfrom()");
+	cleanup(ctx);
+	DEBUG_EXIT("envfrom", "SMFIS_TEMPFAIL");
+	return SMFIS_TEMPFAIL;
     }
 
-    if (!success) {
-	if (protectMkdirWithMutex) {
-	    if (pthread_mutex_lock(&MkdirMutex)) {
-		syslog(LOG_INFO, "Could not lock MkdirMutex");
-		cleanup(ctx);
-		DEBUG_EXIT("envfrom", "SMFIS_TEMPFAIL");
-		return SMFIS_TEMPFAIL;
-	    }
-	}
-
-	for (tries = 0; tries < MAXTRIES; tries++) {
-	    snprintf(buffer, SMALLBUF, "%s/mdefang-%lX-%d",
-		     SpoolDir,
-		     ulnow, tries);
-	    if (!mkdir(buffer, 0750)) {
-		success = 1;
-		break;
-	    }
-
-	    /* Apparently, mkdir on Solaris 8 can fail with EBADF */
-	    if ((errno != EEXIST) && (errno != EBADF)) {
-		break;
-	    }
-	}
-
-	if (protectMkdirWithMutex) {
-	    (void) pthread_mutex_unlock(&MkdirMutex);
-	}
-    }
-
-    /* Could not create temp. directory */
-    if (!success) {
+    snprintf(buffer, SMALLBUF, "%s/mdefang-%s", SpoolDir, data->qid);
+    if (mkdir(buffer, 0750) != 0) {
+        /* Could not create temp. directory */
 	syslog(LOG_WARNING, "%s: Could not create directory %s: %m",
 	       data->qid, buffer);
 	cleanup(ctx);
@@ -821,6 +867,9 @@ envfrom(SMFICTX *ctx, char **from)
     /* Initialize the dynamic buffer for writing COMMANDS file */
     dbuf_init(&dbuf);
 
+    /* Generate an ID */
+    append_mx_command(&dbuf, 'i', gen_mx_id(mxid));
+
     /* Write the sender */
     append_mx_command(&dbuf, 'S', from[0]);
 
@@ -830,6 +879,7 @@ envfrom(SMFICTX *ctx, char **from)
     }
 
     /* Write some macros */
+    // TODO: Just have one array of macros (regular plus additional) and iterate over it
     append_macro_value(&dbuf, ctx, "_");
     append_macro_value(&dbuf, ctx, "auth_authen");
     append_macro_value(&dbuf, ctx, "auth_author");
@@ -1637,7 +1687,8 @@ eom(SMFICTX *ctx)
 	    syslog(LOG_WARNING, "%s: Overlong line in RESULTS file - %d chars (max %d)",
 		   data->qid, (int) (eptr - rptr), SMALLBUF-1);
 	    cleanup(ctx);
-	    r = SMFIS_TEMPFAIL;
+	    MD_SMFI_TRY(set_reply, (ctx, "5", "554", "5.4.0" , "Overlong line in RESULTS file"));
+	    r = SMFIS_REJECT;
 	    goto bail_out;
 	}
 
@@ -2073,7 +2124,6 @@ usage(void)
     fprintf(stderr, "  -C                -- Try very hard to conserve file descriptors\n");
     fprintf(stderr, "  -x string         -- Add string as X-Scanned-By header\n");
     fprintf(stderr, "  -X                -- Do not add X-Scanned-By header\n");
-    fprintf(stderr, "  -M                -- Protect mkdir with mutex\n");
     fprintf(stderr, "  -D                -- Do not become a daemon (stay in foreground)\n");
     fprintf(stderr, "  -S facility       -- Set syslog(3) facility\n");
     fprintf(stderr, "  -a macro          -- Pass additional Sendmail macro\n");
@@ -2237,7 +2287,7 @@ main(int argc, char **argv)
 	    }
 	    break;
 	case 'M':
-	    protectMkdirWithMutex = 1;
+	    /* Ignore.  This once set protectMkdirWithMutex, which has been removed. */
 	    break;
 	case 'X':
 	    if (scan_body) {
